unit LogManager;

interface

uses Classes;

type
  TLinePriority = (lpNone, lpSuccess, lpWarning, lpError, lpDebug,
                   lpDeletedFile, lpStdWork, lpSpaceEarn, lpTime);
  TOnLogChange = procedure(const LineAdded: string; const Priority: TLinePriority) of object;

  TLogManager = class
  private
    FOptions: TObject;
    FNumLine: Integer;
    FOnChange: TOnLogChange;
    FDirName: string;
    FStream: TFileStream;
    procedure SetOnChange(const Value: TOnLogChange);
    procedure SetDirName(const Value: string);
    function GetFileName: string;
  protected
    procedure AddLine(Line: String);
  public
    constructor Create(Options: TObject);

    procedure Add(Line: String; const Priority: TLinePriority = lpNone);
    procedure Detach;

    property DirName: string read FDirName write SetDirName;
    property FileName: string read GetFileName;

    property OnChange: TOnLogChange read FOnChange write SetOnChange;
  end;

const
  cstLogFileName = 'xdelwin.log';

implementation

uses Windows, SysUtils, OptionsManager, resources;


{ TLogManager }

procedure TLogManager.Add(Line: String; const Priority: TLinePriority);
begin
  case Priority of
    lpDebug : begin
      if not ((FOptions as TOptionsManager).Debug and
              (FOptions as TOptionsManager).DebugOptions.DetailsLog) then Exit;
      Line := rsLogDebug + Line;
      if Assigned(FOnChange) then FOnChange(Line, Priority);
    end;
    lpWarning : begin
      Line := rsLogWarning + Line;
      if leWarning in (FOptions as TOptionsManager).Log.Events then AddLine(Line);
      if Assigned(FOnChange) then FOnChange(Line, Priority);
    end;
    lpError   : begin
      Line := rsLogError + Line;
      if leError in (FOptions as TOptionsManager).Log.Events then AddLine(Line);
      if Assigned(FOnChange) then FOnChange(Line, Priority);
    end;
    lpDeletedFile : begin
      Line := rsLogDeleted + Line;
      if leDeletedFile in (FOptions as TOptionsManager).Log.Events then AddLine(Line);
      if Assigned(FOnChange) then FOnChange(Line, lpNone);
    end;
    lpStdWork : begin
      if leStdClean in (FOptions as TOptionsManager).Log.Events then AddLine(Line);
      if Assigned(FOnChange) then FOnChange(Line, lpNone);
    end;
    lpSpaceEarn : begin
      if leSizeFree in (FOptions as TOptionsManager).Log.Events then AddLine(Line);
      if Assigned(FOnChange) then FOnChange(Line, lpSuccess);
    end;
    lpTime : begin
      if leTimeSearch in (FOptions as TOptionsManager).Log.Events then AddLine(Line);
      if Assigned(FOnChange) then FOnChange(Line, lpNone);
    end;
    else begin
      Line := Line;
      if Assigned(FOnChange) then FOnChange(Line, lpNone);
    end;
  end;
end;

procedure TLogManager.AddLine(Line: String);
var H: string;
begin
  if not Assigned(FStream) then begin
    if ((FOptions as TOptionsManager).Log.EraseOldFile and (FNumLine = 0)) or (not FileExists(FileName)) then
      FStream := TFileStream.Create(FileName, fmCreate or fmShareDenyWrite)
    else
      FStream := TFileStream.Create(FileName, fmOpenReadWrite or fmShareDenyWrite);

    FStream.Seek(0, soEnd);
    if FNumLine = 0 then begin
      H := Format(rsLogHeader, [DateTimeToStr(Now)]);
      FStream.Write(PChar(H)^, Length(H));
    end;
  end;
  Line := Line + sLineBreak;
  FStream.Write(PChar(Line)^, Length(Line));
  Inc(FNumLine);
end;

constructor TLogManager.Create(Options: TObject);
begin
  FStream := nil;
  FNumLine := 0;
  FOptions := Options;
end;

procedure TLogManager.Detach;
begin
  FreeAndNil(FStream);
end;

function TLogManager.GetFileName: string;
begin
  result := IncludeTrailingPathDelimiter(FDirName) + cstLogFileName;
end;

procedure TLogManager.SetDirName(const Value: string);
begin
  if FDirName <> Value then begin
    FreeAndNil(FStream);
    if FileExists(FileName) then
      MoveFile(PChar(FileName), PChar(IncludeTrailingPathDelimiter(Value) + cstLogFileName));
    FDirName := Value;
  end;
end;

procedure TLogManager.SetOnChange(const Value: TOnLogChange);
begin
  FOnChange := Value;
end;

end.
