unit SearchObject;

interface

uses Classes, SysUtils, ComCtrls, Contnrs, Hash, FileTools;

type
  TGroup = class
  private
    FMaskList: TStrings;
    FRecurse: boolean;
    FName: ShortString;
    FOnChange: TNotifyEvent;
    FCRC: LongWord;
    procedure SetRecurse(const Value: boolean);
    procedure SetName(const Value: ShortString);
    procedure SetOnChange(const Value: TNotifyEvent);
  protected
    procedure MaskChange(Sender: TObject);
  public
    constructor Create(const Name: ShortString);
    destructor Destroy; override;

    procedure LoadFromStream(const St: TStream);
    procedure SaveToStream(const St: TStream);

    procedure GenerateTreeNode(const TreeNodes: TTreeNodes; const Parent: TTreeNode);
    
    property CRC: LongWord read FCRC;
    property Name: ShortString read FName write SetName;
    property Mask: TStrings read FMaskList;
    property Recurse: boolean read FRecurse write SetRecurse;
    property OnChange: TNotifyEvent read FOnChange write SetOnChange;
  end;

  TGroupList = class(TObjectList)
  private
    FDefaultGroup: TGroup;
    FOnChange: TNotifyEvent;
    procedure AddDefaultGroup;
    procedure GroupChange(Sender: TObject);
    procedure SetOnChange(const Value: TNotifyEvent);
  protected
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
    property OwnsObjects;
  public
    constructor Create;

    procedure LoadFromStream(const St: TStream);
    procedure SaveToStream(const St: TStream);

    procedure CopyTo(GroupList: TGroupList);

    function IndexOf(CRC: LongWord): Integer; overload;
    function IndexOf(Name: String): Integer; overload;
    procedure Clear; override;

    property DefaultGroup: TGroup read FDefaultGroup;
    property OnChange: TNotifyEvent read FOnChange write SetOnChange;
  end;

  TDirectory = class(TObjectList)
  private
    FName: ShortString;
    FGroups: TObjectList;
    FGroupList: TGroupList;
    FOwner: TDirectory;
    procedure SetName(const Value: ShortString);
  protected
    function GetPath: string; virtual;
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;

    property OwnsObjects;
  public
    constructor Create(const Name: string; const GroupList: TGroupList; const Owner: TDirectory = nil);
    destructor Destroy; override;

    procedure LoadFromStream3(const St: TStream);
    procedure LoadFromStream(const St: TStream);
    procedure SaveToStream(const St: TStream);

    procedure Delete(); overload;

    function IndexOf(const St: ShortString): Integer; overload;
    function IndexOf(const CRC: LongWord): Integer; overload;
    procedure AddPath(const Path: string);
    function Browse(const Dir: string; const null: boolean = True): TDirectory;
    procedure GenerateTreeNode(const TreeNodes: TTreeNodes; const Parent: TTreeNode; const Sel: TDirectory = nil);
    procedure BaseDirectoryTo(const List: TStrings);
    procedure Masques(const Msk: TStrings);
    procedure MasquesRecursif(const Msk: TStrings);

    property Path: string read GetPath;
    property Name: ShortString read FName write SetName;
    property Groups: TObjectList read FGroups;
  end;

  TDrive = class(TDirectory)
  private
    FDriveType: TDriveType;
  protected
    function GetPath: String; override;
  public
    constructor Create(const Letter: Char; const GroupList: TGroupList);

    property DriveType: TDriveType read FDriveType;
  end;

  TDrives = class(TObject)
  private
    FDrives: array[1..26] of TDrive;
    FGroupList: TGroupList;
    function GetItem(Index: Integer): TDrive;
  public
    constructor Create(const GroupList: TGroupList);
    destructor Destroy; override;

    procedure Clear;

    procedure GenerateTreeNode(const TreeNodes: TTreeNodes; const Sel: TDirectory = nil);
    function Browse(const Dir: string; const null: boolean = True): TDirectory;
    procedure BaseDirectoryTo(const List: TStrings);
    procedure AddPath(const Path: string);

    procedure LoadFromStream3(const St: TStream);
    procedure LoadFromStream(const St: TStream);
    procedure SaveToStream(const St: TStream);

    procedure CopyTo(Drives: TDrives);

    property Items[Index: Integer]: TDrive read GetItem;
  end;

implementation

uses IniFiles, resources;

function CompareCRC(Item1, Item2: Pointer): Integer;
var Diff      : Int64;
begin
  Diff := Int64(TGroup(Item1).CRC) - Int64(TGroup(Item2).CRC);
  if Diff < 0 then result := -1 else
  if Diff > 0 then result := 1  else
  result := 0;
end;

function CompareString(Item1, Item2: Pointer): Integer;
begin
  result := CompareStr(TDirectory(Item1).FName, TDirectory(Item2).FName);
end;

{ TGroup }

constructor TGroup.Create(const Name: ShortString);
begin
  inherited Create;

  FMaskList := TStringList.Create;
  (FMaskList as TStringList).Sorted := True;
  (FMaskList as TStringList).Duplicates := dupIgnore;
  (FMaskList as TStringList).OnChange := MaskChange;

  FRecurse := True;
  FName := Name;
  FCRC := not CRC32($FFFFFFFF, @Name, Length(Name) + 1);
end;

destructor TGroup.Destroy;
begin
  FMaskList.Free;
  inherited;
end;

procedure TGroup.GenerateTreeNode(const TreeNodes: TTreeNodes;
  const Parent: TTreeNode);
var T, M: TTreeNode;
    I: Integer;
begin
  T := TreeNodes.AddChild(Parent, FName);
  T.ImageIndex := 7;
  T.SelectedIndex := 7;

  for I := 0 to FMaskList.Count - 1 do begin
    M := TreeNodes.AddChild(T, FMaskList[I]);
    M.ImageIndex := 8;
    M.SelectedIndex := 8;
  end;
end;

procedure TGroup.LoadFromStream(const St: TStream);
var StrLength, MaskCount: Byte;
    I, J                : Integer;
    S                   : String;
begin
  St.Read(FCRC, SizeOf(FCRC));
  St.Read(StrLength, SizeOf(StrLength));
  SetLength(FName, StrLength);
  for I := 1 to StrLength do
    St.Read(FName[I], SizeOf(Char));

  St.Read(FRecurse, SizeOf(FRecurse));
  St.Read(MaskCount, SizeOf(MaskCount));

  FMaskList.Clear;
  FMaskList.Capacity := MaskCount;
  for I := 1 to MaskCount do begin
    St.Read(StrLength, SizeOf(StrLength));
    SetLength(S, StrLength);
    for J := 1 to StrLength do
      St.Read(S[J], SizeOf(Char));

    FMaskList.Add(S);
  end;
end;

procedure TGroup.MaskChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TGroup.SaveToStream(const St: TStream);
var B: Byte;
    I: Integer;
    S: ShortString;
begin
  St.Write(FCRC, SizeOf(FCRC));
  St.Write(FName, Length(FName) + 1);
  St.Write(FRecurse, SizeOf(FRecurse));
  B := FMaskList.Count;
  St.Write(B, SizeOf(B));
  for I := 0 to B - 1 do begin
    S := FMaskList[I];
    St.Write(S, Length(S) + 1);
  end;
end;

procedure TGroup.SetName(const Value: ShortString);
begin
  if FName <> Value then begin
    FName := Value;
    if Assigned(FOnChange) then
      FOnChange(Self);
  end;
end;

procedure TGroup.SetOnChange(const Value: TNotifyEvent);
begin
  FOnChange := Value;
end;

procedure TGroup.SetRecurse(const Value: boolean);
begin
  if FRecurse <> Value then begin
    FRecurse := Value;
    if Assigned(FOnChange) then
      FOnChange(Self);
  end;
end;

{ TGroupList }

procedure TGroupList.AddDefaultGroup;
begin
  FDefaultGroup := TGroup.Create(rsMainGroup);
  FDefaultGroup.Recurse := True;
  FDefaultGroup.Mask.Add('*.$$$');
  FDefaultGroup.Mask.Add('*.~??');
  FDefaultGroup.Mask.Add('*.bak');
  FDefaultGroup.Mask.Add('*.old');
  FDefaultGroup.Mask.Add('*.tmp');
  FDefaultGroup.Mask.Add('file*.chk');
  FDefaultGroup.Mask.Add('~$*.*');
  FDefaultGroup.Mask.Add('*.shd');

  Add(FDefaultGroup);
end;

procedure TGroupList.Clear;
begin
  inherited;
  AddDefaultGroup;
end;

procedure TGroupList.CopyTo(GroupList: TGroupList);
var St: TMemoryStream;
begin
  St := TMemoryStream.Create;
  try
    SaveToStream(St);
    St.Position := 0;
    GroupList.LoadFromStream(St);
  finally
    St.Free;
  end;
end;

constructor TGroupList.Create;
begin
  inherited;

  OwnsObjects := True;
  AddDefaultGroup;
end;

procedure TGroupList.GroupChange(Sender: TObject);
begin
  if Assigned(FOnChange) then
    FOnChange(Sender);
end;

function TGroupList.IndexOf(CRC: LongWord): Integer;
var Min, Max, Med: Integer;
begin
  if Count > 0 then begin
    Min := 0; Max := Count - 1;
    repeat
      Med := (Min + Max) div 2;
      if (CRC > TGroup(Items[Med]).CRC) then
        Min := Med + 1
      else
        Max := Med - 1;
    until (TGroup(Items[Med]).CRC = CRC) or (Min > Max);
    if (TGroup(Items[Med]).CRC = CRC) then
      result := Med
    else
      result := -1;
  end else
    result := -1;
end;

function TGroupList.IndexOf(Name: String): Integer;
var I: Integer;
begin
  result := -1;
  for I := 0 to Count - 1 do
    if CompareText(TGroup(Items[I]).FName, Name) = 0 then begin
      result := I;
      break;
    end;
end;

procedure TGroupList.LoadFromStream(const St: TStream);
var I  : Integer;
    Grp: TGroup;
    Cnt: Byte;
begin
  Clear;

  St.Read(Cnt, SizeOf(Cnt));
  for I := 0 to Cnt - 1 do begin
    Grp := TGroup.Create('');
    Grp.LoadFromStream(St);
    Add(Grp);
  end;
end;

procedure TGroupList.Notify(Ptr: Pointer; Action: TListNotification);
begin
  inherited;

  if Action = lnAdded then begin
    TGroup(Ptr).OnChange := GroupChange;
    Sort(CompareCRC);
  end;

  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TGroupList.SaveToStream(const St: TStream);
var I  : Integer;
    Cnt: Byte;
begin
  Cnt := Count - 1;
  St.Write(Cnt, SizeOf(Cnt));
  for I := 0 to Cnt do
    if Items[I] <> FDefaultGroup then
      TGroup(Items[I]).SaveToStream(St);
end;

procedure TGroupList.SetOnChange(const Value: TNotifyEvent);
begin
  FOnChange := Value;
end;

{ TDirectory }

procedure TDirectory.AddPath(const Path: string);
var I: Integer;
    Dir: TDirectory;
    S: String;
begin
  if Path <> '' then begin
    S := Copy(Path, 1, Pos(PathDelim, Path) - 1);
    I := IndexOf(S);
    if I = - 1 then begin
      Dir := TDirectory.Create(S, FGroupList, Self);
      Add(Dir);
    end else begin
      Dir := TDirectory(Items[I]);
    end;
    Dir.AddPath(Copy(Path, Pos(PathDelim, Path) + 1, Length(Path)));
  end;
end;

procedure TDirectory.BaseDirectoryTo(const List: TStrings);
var I: Integer;
begin
  if Groups.Count > 0 then
    List.Add(Path)
  else
    for I := 0 to Count - 1 do
      TDirectory(Items[I]).BaseDirectoryTo(List);
end;

function TDirectory.Browse(const Dir: string; const null: boolean = True): TDirectory;
var I: Integer;
begin
  if Dir <> '' then begin
    I := IndexOf(Copy(Dir, 1, Pos(PathDelim, Dir) - 1));
    if I = - 1 then begin
      if null then
        result := nil
      else
        result := Self;
    end else
      result := TDirectory(Items[I]).Browse(Copy(Dir, Pos(PathDelim, Dir) + 1, Length(Dir)), null);
  end else
    result := Self;
end;

constructor TDirectory.Create(const Name: string; const GroupList: TGroupList; const Owner: TDirectory = nil);
begin
  inherited Create;

  OwnsObjects := True;
  FOwner := Owner;
  FName := Name;
  FGroupList := GroupList;
  FGroups := TObjectList.Create;
  FGroups.OwnsObjects := false;
end;

procedure TDirectory.Delete;
begin
  FOwner.Delete(FOwner.IndexOf(Self));
end;

destructor TDirectory.Destroy;
begin
  FGroups.Free;

  inherited;
end;

procedure TDirectory.GenerateTreeNode(const TreeNodes: TTreeNodes;
  const Parent: TTreeNode; const Sel: TDirectory = nil);
var T: TTreeNode;
    I: Integer;
begin
  T := TreeNodes.AddChild(Parent, FName);
  if Sel = Self then
    T.TreeView.Selected := T;
    
  if not (Self is TDrive) then begin
    if DirExists(Path) then begin
      T.ImageIndex := 5;
      T.SelectedIndex := 6;
    end else begin
      T.ImageIndex := 9;
      T.SelectedIndex := 9;
    end;
  end else begin
    case (Self as TDrive).FDriveType of
      Removable : begin
        T.ImageIndex := 1;
        T.SelectedIndex := 1;
      end;
      Fixed : begin
        T.ImageIndex := 2;
        T.SelectedIndex := 2;
      end;
      Remote : begin
        T.ImageIndex := 4;
        T.SelectedIndex := 4;
      end;
      CDROM : begin
        T.ImageIndex := 3;
        T.SelectedIndex := 3;
      end;
      else begin
        T.ImageIndex := 9;
        T.SelectedIndex := 9;
      end;
    end;
  end;

  for I := 0 to Count - 1 do
    TDirectory(Items[I]).GenerateTreeNode(TreeNodes, T, Sel);

  for I := 0 to Groups.Count - 1 do
    TGroup(Groups[I]).GenerateTreeNode(TreeNodes, T);
end;

function TDirectory.GetPath: string;
begin
  if FOwner <> nil then
    result := FOwner.GetPath + '\' + FName
  else
    result := FName;
end;

function TDirectory.IndexOf(const St: ShortString): Integer;
var Min, Max, Med: Integer;
begin
  if Count > 0 then begin
    Min := 0; Max := Count - 1;
    repeat
      Med := (Min + Max) div 2;
      if CompareStr(St, TDirectory(Items[Med]).FName) > 0 then
        Min := Med + 1
      else
        Max := Med - 1;
    until (CompareStr(St, TDirectory(Items[Med]).FName) = 0) or (Min > Max);
    if CompareStr(St, TDirectory(Items[Med]).FName) = 0 then
      result := Med
    else
      result := -1;
  end else
    result := -1;
end;

function TDirectory.IndexOf(const CRC: LongWord): Integer;
var I: Integer;
begin
  result := -1;
  for I := 0 to FGroups.Count - 1 do
    if TGroup(FGroups[I]).FCRC = CRC then begin
      result := I;
      break;
    end;
end;

procedure TDirectory.LoadFromStream(const St: TStream);
var I, Cnt, Index: Integer;
    StrLength: Byte;
    CRC: LongWord;
    Dir: TDirectory;
begin
  St.Read(StrLength, SizeOf(StrLength));
  SetLength(FName, StrLength);
  for I := 1 to StrLength do
    St.Read(FName[I], SizeOf(Char));

  FGroups.Clear;
  St.Read(Cnt, SizeOf(Cnt));
  for I := 1 to Cnt do begin
    St.Read(CRC, SizeOf(CRC));
    Index := FGroupList.IndexOf(CRC);
    if Index <> -1 then
      FGroups.Add(FGroupList[Index]);
  end;

  Clear;
  St.Read(Cnt, SizeOf(Cnt));
  for I := 1 to Cnt do begin
    Dir := TDirectory.Create('', FGroupList, Self);
    Dir.LoadFromStream(St);
    Add(Dir);
  end;
end;

procedure TDirectory.LoadFromStream3(const St: TStream);
var I, J, Cnt: Integer;
    StrLength: Byte;
    Dir: TDirectory;

    L1, L2: TGroup;
    Chaine: ShortString;
    Recurs: boolean;
begin
  St.Read(StrLength, SizeOf(StrLength));
  SetLength(FName, StrLength);
  for I := 1 to StrLength do
    St.Read(FName[I], SizeOf(Char));

  Clear;
  St.Read(Cnt, SizeOf(Cnt));
  for I := 1 to Cnt do begin
    Dir := TDirectory.Create('', FGroupList, Self);
    Dir.LoadFromStream3(St);
    Add(Dir);
  end;

  FGroups.Clear;
  St.Read(Cnt, SizeOf(Cnt));

  L1 := nil; L2 := nil;
  try
    L1 := TGroup.Create(Format(rsGroupWithDir, [FGroupList.Count + 1]));
    L1.Recurse := True;
    L2 := TGroup.Create(Format(rsGroupWithoutDir, [FGroupList.Count + 1]));
    L2.Recurse := False;

    for I := 1 to Cnt do begin
      St.Read(StrLength, SizeOf(StrLength));
      SetLength(Chaine, StrLength);
      for J := 1 to StrLength do
        St.Read(Chaine[J], SizeOf(Char));
      St.Read(Recurs, SizeOf(Recurs));

      if Recurs then
        L1.FMaskList.Add(Chaine)
      else
        L2.FMaskList.Add(Chaine);
    end;

    if L1.FMaskList.Count > 0 then begin
      FGroups.Add(L1);
      FGroupList.Add(L1);
    end else
      L1.Free;
    if L2.FMaskList.Count > 0 then begin
      FGroups.Add(L2);
      FGroupList.Add(L2);
    end else
      L2.Free;
  except
    L1.Free;
    L2.Free;
  end;
end;

procedure TDirectory.Masques(const Msk: TStrings);
var I: Integer;
begin
  for I := 0 to Groups.Count - 1 do begin
    Msk.AddStrings(TGroup(Groups[I]).FMaskList);
  end;
  if Assigned(FOwner) then
    FOwner.MasquesRecursif(Msk);
end;

procedure TDirectory.MasquesRecursif(const Msk: TStrings);
var I: Integer;
begin
  for I := 0 to Groups.Count - 1 do begin
    if TGroup(Groups[I]).FRecurse then
      Msk.AddStrings(TGroup(Groups[I]).FMaskList);
  end;
  if Assigned(FOwner) then
    FOwner.MasquesRecursif(Msk);
end;

procedure TDirectory.Notify(Ptr: Pointer; Action: TListNotification);
begin
  inherited;
  if Action = lnAdded then begin
    Sort(CompareString);
  end;
end;

procedure TDirectory.SaveToStream(const St: TStream);
var I, Cnt: Integer;
    CRC: LongWord;
begin
  St.Write(FName, Length(FName) + 1);
  Cnt := Groups.Count;
  St.Write(Cnt, SizeOf(Cnt));
  for I := 0 to Cnt - 1 do begin
    CRC := TGroup(Groups[I]).CRC;
    St.Write(CRC, SizeOf(CRC));
  end;

  Cnt := Count;
  St.Write(Cnt, SizeOf(Cnt));
  for I := 0 to Cnt - 1 do
    TDirectory(Items[I]).SaveToStream(St);
end;

procedure TDirectory.SetName(const Value: ShortString);
begin
  if FName <> Value then begin
    FName := Value;
  end;
end;

{ TDrive }

constructor TDrive.Create(const Letter: Char; const GroupList: TGroupList);
begin
  inherited Create(Letter, GroupList);

  FDriveType := FileTools.DriveType(Letter);
  if FDriveType = Fixed then
    Groups.Add(GroupList.DefaultGroup);
end;

function TDrive.GetPath: String;
begin
  result := FName + ':';
end;

{ TDrives }

procedure TDrives.AddPath(const Path: string);
begin
  FDrives[Ord(UpCase(Path[1])) - Ord('A') + 1].AddPath(
    Copy(Path, Pos(PathDelim, Path) + 1, Length(Path))
  );
end;

procedure TDrives.BaseDirectoryTo(const List: TStrings);
var I: Integer;
begin
  for I := 1 to 26 do
    FDrives[I].BaseDirectoryTo(List);
end;

function TDrives.Browse(const Dir: string; const null: boolean = True): TDirectory;
begin
  result := FDrives[Ord(UpCase(Dir[1])) - Ord('A') + 1].Browse(
    Copy(Dir, Pos(PathDelim, Dir) + 1, Length(Dir)), null
  );
end;

procedure TDrives.Clear;
var I: Byte;
begin
  for I := 1 to 26 do begin
    FDrives[I].Free;
    FDrives[I] := TDrive.Create(Chr(Ord('A') - 1 + I), FGroupList);
  end;
end;

procedure TDrives.CopyTo(Drives: TDrives);
var St: TMemoryStream;
begin
  St := TMemoryStream.Create;
  try
    SaveToStream(St);
    St.Position := 0;
    Drives.LoadFromStream(St);
  finally
    St.Free;
  end;
end;

constructor TDrives.Create(const GroupList: TGroupList);
var I: Byte;
begin
  inherited Create;

  FGroupList := GroupList;

  for I := 1 to 26 do
    FDrives[I] := TDrive.Create(Chr(Ord('A') - 1 + I), GroupList);
end;

destructor TDrives.Destroy;
var I: Byte;
begin
  for I := 1 to 26 do
    FDrives[I].Free;
end;

procedure TDrives.GenerateTreeNode(const TreeNodes: TTreeNodes; const Sel: TDirectory = nil);
var I: Integer;
begin
  for I := 1 to 26 do
    if (FDrives[I].DriveType <> no_root) or (FDrives[I].Count > 0) or
       (FDrives[I].Groups.Count > 0) then
    FDrives[I].GenerateTreeNode(TreeNodes, nil, Sel);
end;

function TDrives.GetItem(Index: Integer): TDrive;
begin
  result := FDrives[Index];
end;

procedure TDrives.LoadFromStream(const St: TStream);
var I, Cnt, Index: Byte;
begin
  St.Read(Cnt, SizeOf(Cnt));
  for I := 1 to Cnt do begin
    St.Read(Index, SizeOf(Cnt));
    FDrives[Index].LoadFromStream(St);
  end;
end;

procedure TDrives.LoadFromStream3(const St: TStream);
var Cnt, I: Integer;
begin
  FGroupList.Clear;
  St.Read(Cnt, SizeOf(Cnt));
  for I := 1 to Cnt do
    FDrives[I].LoadFromStream3(St);
end;

procedure TDrives.SaveToStream(const St: TStream);
var I, Cnt: Byte;
    Pos, Pos2: Int64;
begin
  Pos := St.Position;

  Cnt := 0;
  St.Write(Cnt, SizeOf(Cnt));
  for I := 1 to 26 do begin
    if FDrives[I].DriveType <> no_root then begin
      St.Write(I, SizeOf(I));
      FDrives[I].SaveToStream(St);
      Inc(Cnt);
    end;
  end;

  Pos2 := St.Position;
  St.Seek(Pos, soBeginning);
  St.Write(Cnt, SizeOf(Cnt));
  St.Seek(Pos2, soCurrent);
end;

end.
